// Copyright (C) 2003 Nicolas Fleury (nidoizo@yahoo.com)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
#ifndef EDV_IMPLICIT_VALUE_HPP
#define EDV_IMPLICIT_VALUE_HPP

#include <edv/value.hpp>
#include <boost/smart_ptr.hpp>
#include <boost/bind.hpp>

namespace edv {
  // implicit_value is used to implicitly receive one of the following
  // with one parameter:
  // - a value of type
  // - a const value<type>&
  // - a const value<type>* (adoption)
  // - a boost::shared_ptr<value<type> > (sharing ownership)
  template <typename Type>
  class implicit_value : public value<Type>, 
    public boost::signals::trackable
  {
  public:
    typedef Type type;
    implicit_value(type v) : 
      value<Type>(v),
      m_implicit_type(owned), 
      m_owned_value(new concrete_value<type>(v)),
      m_value(*m_owned_value) 
    { 
      init();
    }
    implicit_value(const value<type>& v) : 
      value<Type>(v.get()),
      m_implicit_type(reference), 
      m_value(v)
    { 
      init();
    }
    implicit_value(const value<type>* v) : 
      value<Type>(v->get()),
      m_implicit_type(owned), 
      m_owned_value(v),
      m_value(*v) 
    { 
      init();
    }
    implicit_value(boost::shared_ptr<type> v) :
      value<Type>(v->get()),
      m_implicit_type(shared_ptr),
      m_shared_value(new boost::shared_ptr<type>(v)),
      m_value(*v)
    { 
      init();
    }
    implicit_value(const implicit_value& i) :
      value<Type>(i.get()),
      m_implicit_type(i.m_implicit_type),
      m_value(i.m_value)
    {
      if (i.m_implicit_type == owned)
      {
        m_owned_value = i.m_owned_value;
	i.m_owned_value = 0;
      }
      else if (i.m_implicit_type == shared_ptr)
      {
        m_shared_value = i.m_shared_value;
      }
      init();
    }
    
    virtual ~implicit_value()
    {
      switch (m_implicit_type)
      {
      case owned:
        delete m_owned_value;
        break;
      case shared_ptr:
        delete m_shared_value;
        break;
      }
    }
  private:
    void init()
    {
      m_value.listen_change(
        boost::bind(&implicit_value<type>::on_value_change, this));
    }
    void on_value_change()
    {
      set(m_value.get());
    }
    
    enum
    {
      reference,
      owned,
      shared_ptr
    } m_implicit_type;
    union
    {
      mutable const value<type>* m_owned_value;
      boost::shared_ptr<type>* m_shared_value;
    };
    const value<type>& m_value;
  };
}  // namespace edv

#endif  // EDV_IMPLICIT_VALUE_HPP
