// Copyright (C) 2003 Nicolas Fleury (nidoizo@yahoo.com)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
#ifndef EDV_STRING_VALUES_HPP
#define EDV_STRING_VALUES_HPP

#include <edv/value.hpp>
#include <edv/implicit_value.hpp>
#include <boost/lexical_cast.hpp>

namespace edv {
  // string_convert_value is used to convert a value to another value
  // of any type that has a constructor taking a const char*.
  template <typename PreviousType, typename String = std::string>
  class string_convert_value : public value<String>, 
    public boost::signals::trackable
  {
  public:
    typedef PreviousType previous_type;
    typedef String type;
    
    string_convert_value(implicit_value<previous_type> v) : 
      value<String>(String(boost::lexical_cast<std::string>(v.get()).c_str())),
      m_value(v)
    {
      m_value.listen_change(boost::bind(
        &string_convert_value<previous_type, type>::on_value_change, 
          this));
    }
    
  private:
    void on_value_change()
    {
      set(String(boost::lexical_cast<std::string>(m_value.get()).c_str()));
    }
    
    implicit_value<previous_type> m_value;
  };

  // ostringstream_value is used to create a ostringstream-like value of
  // a type (String) that has a constructor taking a const char*.  This
  // ostringstream-like value can be used different type of values:
  // value<some-type>& (referenced for updates, some-type convertible with
  //   a lexical_cast to String)
  // value<some-type>* (same a reference, but adopted)
  // some-type (used by copy and appended to stream)
  template <typename String = std::string>
  class ostringstream_value : public value<String>, 
    public boost::signals::trackable
  {
  public:
    typedef String type;

    // The helper class is used to receive all non-value types
    // with operator<<.  Since implicit conversions are done at
    // last resort, the operator<<(helper) will be called only for
    // non-value types.
    class helper
    {
    public:
      helper(const helper& h) : m_string(h.m_string) {}
      
      template <typename Type>
      helper(const Type& v) : 
        m_string(new concrete_value<type>(
	  type(boost::lexical_cast<std::string>(v).c_str()))) {}
      
      std::auto_ptr<value<type> > release() const
      {
        return m_string;
      }
    private:
      mutable std::auto_ptr<value<type> > m_string;
    };
    
    ostringstream_value() : value<String>(String("")) {}
    
    template <typename Type>
    ostringstream_value<type>& operator<<(const value<Type>& v)
    {
      insert(new string_convert_value<Type, type>(v));
      return *this;
    }
    
    template <typename Type>
    ostringstream_value<type>& operator<<(const value<Type>* v)
    {
      insert(new string_convert_value<Type, type>(v));
      return *this;
    }
    
    template <typename Value>
    ostringstream_value<type>& operator<<(boost::shared_ptr<Value> v)
    {
      insert(new string_convert_value<typename Value::type, type>(v));
      return *this;
    }
    
    template <typename Value>
    ostringstream_value<type>& operator<<(std::auto_ptr<Value> v)
    {
      insert(new string_convert_value<typename Value::type, type>(v));
      return *this;
    }
    
    ostringstream_value<type>& operator<<(helper v)
    {
      insert(v.release().release());
      return *this;
    }
  private:
    void insert(value<type>* v)
    {
      m_strings.push_back(v);
      v->listen_change(
        boost::bind(&ostringstream_value<type>::update, this));
      update();
    }
    void update()
    {
      std::stringstream str;
      for (typename std::list<value<type>*>::iterator i = m_strings.begin();
        i != m_strings.end(); ++i)
      {
        str << (*i)->get();
      }
      set(type(str.str().c_str()));
    }
    std::list<value<type>*> m_strings;
  };
  
}  // namespace edv

#endif  // EDV_STRING_VALUES_HPP
