// Copyright (C) 2003 Nicolas Fleury (nidoizo@yahoo.com)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
#ifndef EDV_SYNC_VALUE_HPP
#define EDV_SYNC_VALUE_HPP

#include <edv/value.hpp>
#include <cassert>
#include <boost/function.hpp>

namespace edv {
  // The sync_value class is used to make values with bi-directional
  // relations.
  template <typename Type>
  class sync_value : public value<Type>, public boost::signals::trackable
  {
  public:
    typedef Type type;
    sync_value(type v) : value<type>(v) {}
    sync_value() : value<type>(type()) {}
    using value<type>::set;
    sync_value<type>& operator=(type v)
    {
      set(v);
      return *this;
    }
    sync_value<type>& operator=(const value<type>& v)
    {
      set(v.get());
      return *this;
    }
    static void connect(
      sync_value<type>& sync1, sync_value<type>& sync2,
      boost::function<type (type value2)> update1,
      boost::function<type (type value1)> update2)
    {
      assert(&sync1 != &sync2);
      sync1.m_relations[&sync2] = update1;
      sync2.m_relations[&sync1] = update2;
      sync2.set(update2(sync1.get()));
      sync1.m_connections[&sync2] =
      	sync1.listen_change(
	  boost::bind(&sync_value::on_change, &sync2, &sync1));
      sync2.m_connections[&sync1] =
        sync2.listen_change(
	  boost::bind(&sync_value::on_change, &sync1, &sync2));
    }

    static void disconnect(
      sync_value<type>& sync1, sync_value<type>& sync2)
    {
      assert(&sync1 != &sync2);
      sync1.m_connections[&sync2].disconnect();
      sync2.m_connections[&sync1].disconnect();
      sync1.m_connections.erase(&sync2);
      sync2.m_connections.erase(&sync1);
      sync1.m_relations.erase(&sync2);
      sync2.m_relations.erase(&sync1);
    }
  private:
    void on_change(sync_value<type>* v)
    {
      assert(m_relations.find(v) != m_relations.end());
      set(m_relations.find(v)->second(v->get()));
    }
    std::map<sync_value<type>*, boost::function<type (type)> > m_relations;
    std::map<sync_value<type>*, boost::signals::connection> m_connections;
  };

  template <typename Type>
  void connect(
    sync_value<Type>& sync1, sync_value<Type>& sync2,
    boost::function<Type (Type value2)> update1,
    boost::function<Type (Type value1)> update2)
  {
    sync_value<Type>::connect(sync1, sync2, update1, update2);
  }
  
  template <typename Type>
  void disconnect(
    sync_value<Type>& sync1, sync_value<Type>& sync2)
  {
    sync_value<Type>::disconnect(sync1, sync2);
  }
    
}  // namespace edv

#endif  // EDV_SYNC_VALUE_HPP


